use crate::events::is_supported_event;
use bevy::{
    ecs::{component::Component, entity::Entity, system::Commands, world::World},
    hierarchy::BuildChildren,
    prelude::default,
    render::color::Color,
    text::{Text, TextStyle},
    ui::{
        node_bundles::{NodeBundle, TextBundle},
        *,
    },
    utils::{EntityHashMap, HashMap},
};
use dioxus::core::{
    BorrowedAttributeValue, ElementId, Mutation, Mutations, Template, TemplateAttribute,
    TemplateNode,
};

pub fn apply_mutations(
    mutations: Mutations,
    parent_to_children: &mut HashMap<(Entity, u8), Entity>,
    children_to_parent: &mut EntityHashMap<Entity, Entity>,
    element_id_to_bevy_ui_entity: &mut HashMap<ElementId, Entity>,
    bevy_ui_entity_to_element_id: &mut EntityHashMap<Entity, ElementId>,
    templates: &mut HashMap<String, BevyTemplate>,
    root_entity: Entity,
    commands: &mut Commands,
) {
    for new_template in mutations.templates {
        templates.insert(
            new_template.name.to_owned(),
            BevyTemplate::from_dioxus(&new_template),
        );
    }

    element_id_to_bevy_ui_entity.insert(ElementId(0), root_entity);
    bevy_ui_entity_to_element_id.insert(root_entity, ElementId(0));
    let mut stack = vec![root_entity];

    for edit in mutations.edits {
        match edit {
            Mutation::AppendChildren { id, m } => {
                let mut parent = commands.entity(element_id_to_bevy_ui_entity[&id]);
                let parent_existing_child_count = parent_to_children
                    .keys()
                    .filter(|(p, _)| *p == parent.id())
                    .count();
                for (i, child) in stack.drain((stack.len() - m)..).enumerate() {
                    parent.add_child(child);
                    parent_to_children.insert(
                        (parent.id(), (parent_existing_child_count + i + 1) as u8),
                        child,
                    );
                }
            }
            Mutation::AssignId { path, id } => todo!(),
            Mutation::CreatePlaceholder { id } => {
                let entity = commands
                    .spawn((NodeBundle::default(), BackgroundColor::default()))
                    .id();
                element_id_to_bevy_ui_entity.insert(id, entity);
                bevy_ui_entity_to_element_id.insert(entity, id);
                stack.push(entity);
            }
            Mutation::CreateTextNode { value, id } => {
                let entity = BevyTemplateNode::from_dioxus(&TemplateNode::Text { text: value })
                    .spawn(commands, parent_to_children, children_to_parent);
                element_id_to_bevy_ui_entity.insert(id, entity);
                bevy_ui_entity_to_element_id.insert(entity, id);
                stack.push(entity);
            }
            Mutation::HydrateText { path, value, id } => {
                let mut entity = *stack.last().unwrap();
                for index in path {
                    entity = parent_to_children[&(entity, *index)];
                }
                commands
                    .entity(entity)
                    .insert(Text::from_section(value, TextStyle::default()));
                element_id_to_bevy_ui_entity.insert(id, entity);
                bevy_ui_entity_to_element_id.insert(entity, id);
            }
            Mutation::LoadTemplate { name, index, id } => {
                let entity = templates[name].roots[index].spawn(
                    commands,
                    parent_to_children,
                    children_to_parent,
                );
                element_id_to_bevy_ui_entity.insert(id, entity);
                bevy_ui_entity_to_element_id.insert(entity, id);
                stack.push(entity);
            }
            Mutation::ReplaceWith { id, m } => todo!(),
            Mutation::ReplacePlaceholder { path, m } => todo!(),
            Mutation::InsertAfter { id, m } => todo!(),
            Mutation::InsertBefore { id, m } => todo!(),
            Mutation::SetAttribute {
                name,
                value,
                id,
                ns: _,
            } => {
                let entity = element_id_to_bevy_ui_entity[&id];
                // TODO: The rest of Style
                match (name, value) {
                    ("background-color", BorrowedAttributeValue::Text(hex)) => {
                        let color = Color::hex(hex).expect(&format!(
                            "Encountered unsupported bevy_dioxus background-color `{hex}`."
                        ));
                        commands.add(modify_component_command::<BackgroundColor, _>(
                            entity,
                            move |background_color| background_color.0 = color,
                        ));
                    }
                    (name, value) => {
                        panic!("Encountered unsupported bevy_dioxus attribute `{name}: {value:?}`.")
                    }
                }
            }
            Mutation::SetText { value, id } => {
                commands
                    .entity(element_id_to_bevy_ui_entity[&id])
                    .insert(Text::from_section(value, TextStyle::default()));
            }
            Mutation::NewEventListener { name, id: _ } => {
                if !is_supported_event(name) {
                    panic!("Encountered unsupported bevy_dioxus event `{name}`.");
                }
            }
            Mutation::RemoveEventListener { .. } => {}
            Mutation::Remove { id } => todo!(),
            Mutation::PushRoot { id } => stack.push(element_id_to_bevy_ui_entity[&id]),
        }
    }
}

pub struct BevyTemplate {
    roots: Box<[BevyTemplateNode]>,
}

enum BevyTemplateNode {
    Node {
        style: (Style, BackgroundColor),
        children: Box<[Self]>,
    },
    TextNode(Text),
}

impl BevyTemplate {
    fn from_dioxus(template: &Template) -> Self {
        Self {
            roots: template
                .roots
                .iter()
                .map(BevyTemplateNode::from_dioxus)
                .collect(),
        }
    }
}

impl BevyTemplateNode {
    fn from_dioxus(node: &TemplateNode) -> Self {
        match node {
            TemplateNode::Element {
                tag,
                namespace: _,
                attrs,
                children,
            } => {
                if *tag != "div" {
                    panic!(
                        "Encountered unsupported bevy_dioxus tag `{tag}`. Only `div` is supported."
                    );
                }
                Self::Node {
                    style: parse_style_attributes(attrs),
                    children: children.iter().map(Self::from_dioxus).collect(),
                }
            }
            TemplateNode::Text { text } => {
                Self::TextNode(Text::from_section(*text, TextStyle::default()))
            }
            TemplateNode::Dynamic { id: _ } => Self::Node {
                style: Default::default(),
                children: Box::new([]),
            },
            TemplateNode::DynamicText { id: _ } => {
                Self::TextNode(Text::from_section("", TextStyle::default()))
            }
        }
    }

    fn spawn(
        &self,
        commands: &mut Commands,
        parent_to_children: &mut HashMap<(Entity, u8), Entity>,
        children_to_parent: &mut EntityHashMap<Entity, Entity>,
    ) -> Entity {
        match self {
            BevyTemplateNode::Node {
                style: (style, background_color),
                children,
            } => {
                // TODO: Can probably use with_children() instead
                let children = children
                    .iter()
                    .map(|child| child.spawn(commands, parent_to_children, children_to_parent))
                    .collect::<Box<[_]>>();
                let parent = commands
                    .spawn((
                        NodeBundle {
                            style: style.clone(),
                            ..default()
                        },
                        background_color.clone(),
                    ))
                    .push_children(&children)
                    .id();
                for (i, child) in children.iter().enumerate() {
                    parent_to_children.insert((parent, i as u8), *child);
                    children_to_parent.insert(*child, parent);
                }
                parent
            }
            Self::TextNode(text) => commands
                .spawn(TextBundle {
                    text: text.clone(),
                    ..default()
                })
                .id(),
        }
    }
}

fn parse_style_attributes(attributes: &[TemplateAttribute]) -> (Style, BackgroundColor) {
    let mut style = Style::default();
    let mut background_color = BackgroundColor::default();
    for attribute in attributes {
        if let TemplateAttribute::Static {
            name,
            value,
            namespace: _,
        } = attribute
        {
            // TODO: The rest of Style
            match (*name, *value) {
                ("display", "flex") => style.display = Display::Flex,
                ("display", "grid") => style.display = Display::Grid,
                ("display", "none") => style.display = Display::None,
                ("position", "relative") => style.position_type = PositionType::Relative,
                ("position", "absolute") => style.position_type = PositionType::Absolute,
                ("flex-direction", "column") => style.flex_direction = FlexDirection::Column,
                ("background-color", hex) => {
                    background_color.0 = Color::hex(hex).expect(&format!(
                        "Encountered unsupported bevy_dioxus background-color `{hex}`."
                    ))
                }
                _ => panic!("Encountered unsupported bevy_dioxus attribute `{name}: {value}`."),
            }
        }
    }
    (style, background_color)
}

fn modify_component_command<C, F>(entity: Entity, f: F) -> impl FnOnce(&mut World) + Send + 'static
where
    C: Component,
    F: FnOnce(&mut C) + Send + 'static,
{
    move |world: &mut World| (f)(&mut world.entity_mut(entity).get_mut::<C>().unwrap())
}
